import os
import urllib

import requests
from tqdm import tqdm

api_url = 'https://api.github.com'
raw_url = 'https://raw.githubusercontent.com'
headers = {}
multipleRepositories = False


class GitHubAPIException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


def process_github(owner, repository=None, branch=None, token=None):
    """
    Initiates the process of getting the files from GitHub.

    Args:
        owner (str): The GitHub user.
        repository (str): The repository to search.
        branch (str): The branch to search.
        token (str): The GitHub token.

    Returns:
        dict: The result of getting the files.
    """
    global multipleRepositories
    TOKEN = os.getenv("GITHUB_TOKEN", default=None)
    if token != None:
        TOKEN = token

    if TOKEN != None:
        headers['Authorization'] = f'Bearer {TOKEN}'

    files = {}
    if repository == None:
        multipleRepositories = True
        files = get_files_from_repositories(owner)
    else:
        print(f'Searching repository {owner}/{repository}:')
        files = get_files_from_repository(owner, repository, branch)
        if files != None:
            files = {"repositories": [files]}

    if files == None:
        print("\nNo dataset files found")
        return None

    return files


def get_files_from_repositories(owner):
    """
    Gets the files from all repositories of the GitHub user.

    Args:
        owner (str): The GitHub user.

    Returns:
        dict: The files from all repositories of the GitHub user.
    """
    print(f'Searching repositories for {owner}:')

    json_repos = get_repositories_from_api(owner)
    if json_repos == None:
        return None

    print(f'\t{len(json_repos)} public repositories found\n')

    result = {"repositories": []}

    number_of_repos = len(json_repos)
    remaining = get_rate_limit()
    if remaining <= number_of_repos:
        number_of_repos = remaining
        json_repos = json_repos[:number_of_repos]
        print(
            f'Warning: Only {number_of_repos} repositories will be analyzed, because the GitHub API rate limit has been exceeded.')

    pbar = tqdm(json_repos,
                desc="Reading repositories", ncols=200, unit=" repo", ascii=' █', bar_format="Reading repository {n_fmt}/{total_fmt} |{bar:20}| r:{desc}")
    for repository in pbar:
        pbar.set_description(repository["name"])
        result_repository = get_files_from_repository(
            owner, repository["name"], repository["default_branch"])

        if result_repository == None:
            continue

        result["repositories"].append(result_repository)

    if len(result["repositories"]) == 0:
        return None

    return result


def get_repositories_from_api(owner):
    """
    Gets the repositories from the GitHub API.

    Args:
        owner (str): The GitHub user.

    Returns:
        dict: The repositories from the GitHub API.
    """
    json_repos = {}
    per_page = 100
    count = per_page
    page = 1
    while count == per_page:
        response = requests.get(
            f'{api_url}/users/{owner}/repos?page={page}&per_page={per_page}', headers=headers)
        if not response.ok:
            error = response.json()
            error_message = error.get('message')
            if error_message == "Not Found":
                raise GitHubAPIException(
                    f'Error: Github User or Organization {error_message}')
            elif error_message == "Bad credentials":
                raise GitHubAPIException("Error: Bad credentials")
            elif "API rate limit exceeded" in error_message:
                if page == 1:   # If the first request has already exceeded the rate limit, we can't continue
                    print(
                        "API rate limit exceeded, could not analyze GitHub user or organization.")
                    return None
                else:           # If the rate limit has been exceeded after the first request, we can continue with the repositories already found
                    print(
                        "API rate limit exceeded, not all repositories will be analyzed.")
                    break

        if page == 1:
            json_repos = response.json()
        else:
            json_repos.extend(response.json())
        page += 1
        count = len(response.json())

    return json_repos


def get_rate_limit():
    """
    Gets the rate limit from the GitHub API.

    Returns:
        int: The rate limit from the GitHub API.
    """
    response = requests.get(f'{api_url}/rate_limit', headers=headers)
    if not response.ok:
        error = response.json()
        error_message = error.get('message')
        if error_message == "Bad credentials":
            raise GitHubAPIException("Error: Bad credentials")

    json_data = response.json()
    return json_data["rate"]["remaining"]


def get_files_from_repository(owner, repository, branch=None):
    """
    Gets the files from the GitHub repository.

    Args:
        owner (str): The GitHub user.
        repository (str): The repository to search.
        branch (str): The branch to search.

    Returns:
        dict: The files from the GitHub repository.
    """
    # In case the branch is not specified, we need to get the default branch
    if branch == None:
        branch = get_default_branch_of_repository(owner, repository)

    response = requests.get(
        f'{api_url}/repos/{owner}/{repository}/git/trees/{branch}?recursive=1', headers=headers)
    if not response.ok:
        error = response.json()
        error_message = error.get('message')
        if error_message == "Not Found":
            if multipleRepositories:
                return None
            else:
                raise GitHubAPIException(f'Error: Repository {error_message}')
        elif error_message == "Bad credentials":
            raise GitHubAPIException("Error: Bad credentials")
        elif error_message == "Git Repository is empty.":
            return None
        elif "API rate limit exceeded" in error_message:
            print("API rate limit exceeded")
            return None

    json_files = response.json()

    result_repository = file_selector(json_files, owner, repository)

    return result_repository


def get_default_branch_of_repository(owner, repository):
    """
    Gets the default branch of the GitHub repository.

    Args:
        owner (str): The GitHub user.
        repository (str): The repository to search.

    Returns:
        str: The default branch of the GitHub repository.
    """
    response = requests.get(
        f'{api_url}/repos/{owner}/{repository}', headers=headers)
    if not response.ok:
        error = response.json()
        error_message = error.get('message')
        if error_message == "Not Found":
            if multipleRepositories:
                return None
            else:
                raise GitHubAPIException(
                    f'Error: Repository {error_message}')
        elif error_message == "Bad credentials":
            raise GitHubAPIException("Error: Bad credentials")
        elif "API rate limit exceeded" in error_message:
            print(
                "API rate limit exceeded, not all repositories have been analyzed.")
            return None

    json_repo = response.json()
    return json_repo["default_branch"]


def file_selector(json_files, owner, repository):
    """
    Selects the data sets files from the GitHub repository.

    Args:
        json_files (dict): The files from the GitHub repository.
        owner (str): The GitHub user.
        repository (str): The repository.

    Returns:
        dict: The data sets files from the GitHub repository.
    """
    result_repository = {"name": repository, "files": []}

    extensions = [".csv", ".tsv", ".xlsx", "xlsm", "xltx",
                  "xltm", ".json", ".jsonl"]

    for file in json_files["tree"]:
        if (file["type"] == "blob"):

            if any(file["path"].endswith(ext) for ext in extensions):
                result_repository["files"].append(
                    urllib.parse.quote(f'{raw_url}/{owner}/{repository}/master/{file["path"]}', safe=':/.'))

    # Only return the result if there are files of any type
    if len(result_repository["files"]) == 0:
        return None

    return result_repository



import datetime
import json
import os
import subprocess
import sys
import tempfile


def show_result_as_text(data, name, deep_search=False):
    """
    Shows the result as text.

    Args:
        data (dict): The data to show.

    Returns:
        None
    """
    generate_report = ask_generate_report()
    if not generate_report:
        return
    data = json.dumps(data, indent=4, sort_keys=True)

    # Write result to temp file and open it
    prefix = f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}_{name}_{"deep-search_" if deep_search else ""}'
    suffix = '.json'
    with tempfile.NamedTemporaryFile(prefix=prefix, suffix=suffix, mode='w', delete=False) as f:
        f.write(data)
        f.close()

    if os.name == 'nt':  # For Windows
        subprocess.Popen(["notepad.exe", f.name])
    elif os.name == 'posix':  # For Linux and macOS
        opener = 'open' if sys.platform == 'darwin' else 'xdg-open'
        subprocess.Popen([opener, f.name])
    else:
        raise OSError('Unsupported operating system')


def ask_generate_report():
    """
    Asks the user if he wants to generate a report.

    Args:
        None.

    Returns:
        bool: True if the user wants to generate a report, False otherwise.
    """
    ask_generate = input("\nDo you want to generate a report? (yes/no): ")
    while ask_generate.lower() not in ("yes", "no"):
        ask_generate = input("Please enter either 'yes' or 'no': ")
    if ask_generate.lower() == "yes":
        return True

    return False



import os

import pandas as pd


def clean_csv_file(file, replace_file=False):
    """
    Cleans csv file.

    Args:
        files (dict): The csv file to clean.
        replace_files (bool): If true, the file will be replaced by the clean one.

    Returns:
        None.
    """
    filename = file["name"]
    positive_headers = file["positive_headers"] if "positive_headers" in file else [
    ]
    positive_columns = file["positive_columns"] if "positive_columns" in file else dict(
    )

    try:
        data = pd.read_csv(filename, comment='#', sep=None,
                           engine='python', skip_blank_lines=True, dtype=str)
    except Exception as e:
        error = {"file": filename, "error": str(e)}

    columns = data.columns.values
    for column in columns:
        if column in positive_headers or column in list(positive_columns.keys()):
            data = data.drop(column, axis=1)

    base_filename, extension = os.path.splitext(filename)

    if not replace_file:
        base_filename = base_filename + "_clean"

    new_filename = base_filename + extension
    data.to_csv(new_filename, index=False)



import argparse
import sys


class CustomHelpFormatter(argparse.HelpFormatter):
    def _format_action_invocation(self, action):
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)

        # Align the args_string with the maximum width
        option_strings = (', '.join(action.option_strings)).ljust(17)

        return option_strings + ' ' + args_string

    def format_help(self):
        help_text = super().format_help()
        return help_text + '\n'  # Append an empty line at the end

    def __init__(self, prog):
        super().__init__(prog, max_help_position=40, width=100)


class CustomArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        self.print_usage(sys.stderr)
        error_message = message + '\n'  # Add an empty line after the error message
        self.exit(2, '%s: error: %s\n' % (self.prog, error_message))



import re


def analize_columns(data, columns):
    """
    Analyzes the columns of a file.

    Args:
        data (list): The data of the file.
        headers (list): The headers of the file.

    Returns:
        list: The columns that may have sensitive information.
    """
    positive_columns = {}
    number_of_rows = len(data)
    ratio = number_of_rows * 0.1

    for column in columns:
        positive_fields = set()

        count = 0
        for index, value in data[column].items():
            result_field = analize_field(str(value))
            if result_field != None:
                count += 1
                positive_fields.add(result_field)

            if count >= ratio:
                break

        if len(positive_fields) != 0:
            positive_columns[column] = list(positive_fields)

    return positive_columns


def analize_field(field):
    """
    Analyzes a field.

    Args:
        field (str): The field to analyze.

    Returns:
        str: The type of sensitive information that the field may have.
    """
    email_pattern = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
    iban_pattern = r'^[a-zA-Z]{2}\d{2}[a-zA-Z0-9]{4}\d{7}([a-zA-Z0-9]?){0,16}$'
    phone_pattern_optional_prefix = r'^(\+\d{1,3})?(\d{4,15})$'
    phone_pattern_with_prefix = r'^\+\d{1,3}\d{4,15}$'
    dni_pattern = r'^[0-9]{8}[TRWAGMYFPDXBNJZSQVHLCKE]$'

    patterns = {"email": email_pattern, "iban": iban_pattern,
                "phone": phone_pattern_with_prefix, "dni": dni_pattern}

    for type, pattern in patterns.items():
        if re.match(pattern, field):
            return type

    return None



from cleaner import (csv_cleaner, excel_cleaner, json_cleaner, show_matches,
                     tsv_cleaner)


def process_cleaner(files):
    generate_clean_files = input(
        "\nDo you want to generate clean files? (yes/no): ")
    while generate_clean_files.lower() not in ("yes", "no"):
        generate_clean_files = input("Please enter either 'yes' or 'no': ")
    if generate_clean_files.lower() == "yes":
        clean_files(files)


def clean_files(files):
    """
    Initiates the process of cleaning the files.

    Args:
        files (dict): The files to clean.
        replace_files (bool): If true, the files will be replaced by the clean ones.

    Returns:
        None.
    """
    repository = files["repositories"][0]

    for file in repository["files"]:
        clean_file(file)


def clean_file(file):
    """
    Cleans the file.

    Args:
        file (str): The file to clean.

    Returns:
        None.
    """
    print("\nCleaning file: " + file["name"])

    matches = ""

    basic_extensions = [".csv", ".tsv", ".json", ".jsonl"]
    if any(file["name"].endswith(ext) for ext in basic_extensions):
        matches = show_matches.get_matches(file)

    if file["name"].endswith('.xlsx') or file["name"].endswith('.xls'):
        matches = show_matches.get_matches_excel(file)

    if file["name"].endswith('.mdb') or file["name"].endswith('.accdb'):
        matches = show_matches.get_matches_access(file)

    print(matches)

    clean, replace = ask_clean_file()

    if clean:
        if file["name"].endswith('.csv'):
            csv_cleaner.clean_csv_file(file, replace)

        if file["name"].endswith('.tsv'):
            tsv_cleaner.clean_tsv_file(file, replace)

        excel_extensions = [".xlsx", ".xlsm", ".xltx", ".xltm"]
        if any(file["name"].endswith(ext) for ext in excel_extensions):
            excel_cleaner.clean_excel_file(file, replace)

        if file["name"].endswith('.json') or file["name"].endswith('.jsonl'):
            json_cleaner.clean_json_file(file, replace)


def ask_clean_file():
    """
    Asks the user if he wants to clean the file.

    Args:
        None.

    Returns:
        bool: True if the user wants to clean the file, False otherwise.
        bool: True if the user wants to replace the file, False otherwise.
    """
    ask_clean = input("Do you want to clean this file? (yes/no): ")
    while ask_clean.lower() not in ("yes", "no"):
        ask_clean = input("Please enter either 'yes' or 'no': ")
    if ask_clean.lower() == "yes":

        replace_file = input(
            "Do you want to replace the file? (yes/no): ")
        while replace_file.lower() not in ("yes", "no"):
            replace_file = input("Please enter either 'yes' or 'no': ")
        if replace_file.lower() == "yes":

            replace_file_2 = input(
                "Are you sure you want to replace the file? (yes/no): ")
            while replace_file_2.lower() not in ("yes", "no"):
                replace_file_2 = input(
                    "Please enter either 'yes' or 'no': ")
            if replace_file_2.lower() == "yes":
                return True, True
            else:
                return True, False
        else:
            return True, False
    else:
        return False, False



# Ignore warnings
import warnings

import pandas as pd

from reader import columns_reader, headers_reader

warnings.simplefilter(action='ignore', category=FutureWarning)


def read_json_file(file, deep_search=False):
    """
    Analyzes the json file.

    Args:
        files (str): The file to analyze.
        deep_search (bool): If true, the content of the files will be analyzed.

    Returns:
        dict: The result of analyzing the json file.
    """
    try:
        data = pd.read_json(file, lines=file.endswith(".jsonl"), dtype=str)
    except Exception as e:
        error = {"file": file, "error": str(e)}
        return None, None

    result_file = {"name": file}

    headers = data.columns.values
    result_headers = headers_reader.analize_headers(headers)

    # Only show headers that have errors
    if len(result_headers) != 0:
        result_file["positive_headers"] = result_headers

    # Only analyze columns if deep_search is enabled
    if deep_search:
        result_columns = columns_reader.analize_columns(data, headers)

        # Only show columns that have errors
        if len(result_columns) != 0:
            result_file["positive_columns"] = result_columns

    if len(result_file) == 1:
        return None, None

    return result_file, None



import os


def process_local(directory=None):
    """
    Initiates the process of getting the files from the local repository.

    Args:
        directory (str): The directory to search.

    Returns:
        dict: The result of getting the files.
    """
    files = get_files_recursively(directory)

    if files == None:
        print("\nNo dataset files found")
        return None

    return files


def get_files_recursively(directory=None):
    """
    Gets the files from the local directory.

    Args:
        directory (str): The directory to search.

    Returns:
        dict: The files from the local directory.
    """
    if directory == None:
        directory = "./"

    result = {"repositories": [{"name": "local", "files": []}]}

    extensions = [".csv", ".tsv", ".xlsx", "xlsm", "xltx",
                  "xltm", ".json", ".jsonl"]

    print(f'Searching directory {directory}:')

    for root, _, files in os.walk(directory):
        for filename in files:

            if any(filename.endswith(ext) for ext in extensions):
                filepath = os.path.join(root, filename)
                result["repositories"][0]["files"].append(filepath)

    # Only return the result if there are files of any type
    if len(result["repositories"][0]["files"]) == 0:
        return None

    return result



def get_matches(file):
    matches = ""

    if "positive_headers" in file:
        for positive_header in file['positive_headers']:
            matches = f"{matches}\tHeader {positive_header} may contain sensible data\n"

    if "positive_columns" in file:
        for positive_column, positive_fields in file['positive_columns'].items():
            matches = f"{matches}\tColumn: {positive_column}:\tDetected fields: {positive_fields}\n"

    return matches


def get_matches_excel(file):
    matches = ""

    if "positive_sheets" in file:
        for positive_sheet in file['positive_sheets']:
            matches = f"{matches}\tSheet {positive_sheet['name']}:\n"

            if "positive_headers" in positive_sheet:
                for positive_header in positive_sheet['positive_headers']:
                    matches = f"{matches}\t\tHeader {positive_header} may contain sensible data\n"

            if "positive_columns" in positive_sheet:
                for positive_column, positive_fields in positive_sheet['positive_columns'].items():
                    matches = f"{matches}\t\tColumn: {positive_column}:\tDetected fields: {positive_fields}\n"
    return matches


def get_matches_access(file):
    matches = ""

    if "positive_tables" in file:
        for positive_table in file['positive_tables']:
            matches = f"{matches}\tTable {positive_table['name']}:\n"

            if "positive_headers" in positive_table:
                for positive_header in positive_table['positive_headers']:
                    matches = f"{matches}\t\tHeader {positive_header} may contain sensible data\n"

            if "positive_columns" in positive_table:
                for positive_column, positive_fields in positive_table['positive_columns'].items():
                    matches = f"{matches}\t\tColumn: {positive_column}:\tDetected fields: {positive_fields}\n"
    return matches



import os

import pandas as pd


def clean_excel_file(file, replace_file=False):
    """
    Cleans excel files.

    Args:
        files (dict): The excel file to clean.
        replace_files (bool): If true, the file will be replaced by the clean one.

    Returns:
        None.
    """
    filename = file["name"]
    positive_sheets = file["positive_sheets"] if "positive_sheets" in file else [
    ]

    try:
        data = pd.ExcelFile(filename, engine='openpyxl')
    except Exception as e:
        error = {"file": filename, "error": str(e)}

    base_filename, extension = os.path.splitext(filename)

    if not replace_file:
        base_filename = base_filename + "_clean"

    new_filename = base_filename + extension

    writer = pd.ExcelWriter(new_filename, engine='xlsxwriter')

    for sheet_name in data.sheet_names:
        sheet_data = data.parse(sheet_name, dtype=str)

        positive_sheet = next(
            (sheet for sheet in positive_sheets if sheet["name"] == sheet_name), None)

        if positive_sheet != None:
            positive_headers = positive_sheet["positive_headers"] if "positive_headers" in positive_sheet else [
            ]
            positive_columns = positive_sheet["positive_columns"] if "positive_columns" in positive_sheet else dict(
            )

            columns = sheet_data.columns.values
            for column in columns:
                if column in positive_headers or column in list(positive_columns.keys()):
                    sheet_data = sheet_data.drop(column, axis=1)

        sheet_data.to_excel(writer, sheet_name=sheet_name, index=False)

    writer.close()



from tqdm import tqdm

from reader import csv_reader, excel_reader, json_reader, tsv_reader


def process_files(files, deep_search=False, wide_search=False):
    """
    Initiates the process of analyzing the files.

    Args:
        files (dict): The files to analyze.
        deep_search (bool): If true, the content of the files will be analyzed.
        wide_search (bool): If true, all the tables or sheets will be analyzed.

    Returns:
        dict: The result of analyzing the files.
    """
    repositories = files["repositories"]
    number_of_repositories = len(repositories)

    if number_of_repositories > 1:
        print(f"\t{number_of_repositories} repositories found to have dataset files")

    result = {"repositories": []}

    for index in range(number_of_repositories):
        repository = repositories[index]

        if number_of_repositories > 1:
            print(
                f"\nAnalyzing repository ({index+1}/{number_of_repositories}) {repository['name']}:")
        else:
            print(f"\nAnalyzing repository {repository['name']}:")

        result_repository = {"name": repository["name"]}

        result_files = []
        result_errors = []

        pbar = tqdm(repository["files"],
                    desc="Analyzing files", ncols=300, unit=" file", ascii=' █', bar_format="\tAnalyzing file {n_fmt}/{total_fmt} |{bar:20}| f:{desc}")
        for file in pbar:
            pbar.set_description(file[-50:])

            result_file, result_error = read_file(
                file, deep_search, wide_search)
            if result_file != None:
                result_files.append(result_file)
            if result_error != None:
                result_errors.append(result_error)

        if len(result_files) != 0:
            result_repository["files"] = result_files
        if len(result_errors) != 0:
            result_repository["errors"] = result_errors

        # Only add repository if it has files with errors
        if len(result_repository) != 1:
            result["repositories"].append(result_repository)

    if len(result['repositories']) == 0:
        print("\nYour files are clean!")
        return None

    return result


def read_file(file, deep_search=False, wide_search=False):
    """
    Analyzes the file.

    Args:
        file (str): The file to analyze.
        deep_search (bool): If true, the content of the files will be analyzed.
        wide_search (bool): If true, all the tables or sheets will be analyzed.

    Returns:
        dict: The result of analyzing the file.
    """
    if file.endswith('.csv'):
        result_file, result_error = csv_reader.read_csv_file(file, deep_search)
        return result_file, result_error

    if file.endswith('.tsv'):
        result_file, result_error = tsv_reader.read_tsv_file(file, deep_search)
        return result_file, result_error

    excel_extensions = [".xlsx", "xlsm", "xltx", "xltm"]
    if any(file.endswith(ext) for ext in excel_extensions):
        result_file, result_error = excel_reader.read_excel_file(
            file, deep_search, wide_search)
        return result_file, result_error

    if file.endswith('.json') or file.endswith('.jsonl'):
        result_file, result_error = json_reader.read_json_file(
            file, deep_search)
        return result_file, result_error

    return None, None



import itertools


def analize_headers(headers):
    """
    Analyzes the headers of a file.

    Args:
        headers (list): The headers of the file.

    Returns:
        list: The headers that may have sensitive information.
    """
    terms = ["email", "phone", "mobile", "iban", "account", "sha", "gpg", "socialsecurity",
             "creditcard", "debitcard", "card", "name", "surname", "lastname", "firstname", "dni",
             "license", "licenses", "lecenseplates", "ip", "ips", "address", "addresses", "gps",
             "coordinate", "coordinates", "location", "password", "latitud", "latitude", "longitud",
             "longitude", "passwords", "secret", "secrets", "key", "hash"]
    suffixes = ["number", "value", "key"]

    combinations = []
    for term, suffix in itertools.product(terms, suffixes):
        combinations.append(term)
        combinations.append(term + ' ' + suffix)
        combinations.append(term + suffix)

    positive_headers = []
    for header in headers:
        header_str = str(header)
        clean_eader = header_str.strip().lower().replace("_", " ").replace("-", " ")
        if clean_eader in combinations:
            positive_headers.append(header)

    return positive_headers



import sys
from importlib.metadata import version

from sensitibot import custom_formatters


def parse_args(args=None):
    parser = custom_formatters.CustomArgumentParser(
        prog='sensitibot',
        description='SensitiBot is a tool to analyze datasets for sensitive information.',
        formatter_class=custom_formatters.CustomHelpFormatter
    )

    subparsers = parser.add_subparsers(dest='command')
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s ' + version('SensitiBot'))

    # sensitibot github
    github_parser = subparsers.add_parser(
        'github', formatter_class=custom_formatters.CustomHelpFormatter, help='analyze GitHub repositories')
    github_parser.add_argument('user', type=str, metavar='USER',
                               help='the GitHub user or organization to analyze')
    github_parser.add_argument('-r', '--repository', metavar='REPO',
                               help='analyze a specific repository')
    github_parser.add_argument('-b', '--branch', metavar='BRANCH',
                               help='analyze a specific branch (only if repository is specified)')
    github_parser.add_argument('-t', '--token', metavar='TOKEN',
                               help='the GitHub token to use for authentication')
    github_parser.add_argument('--deep-search', action='store_true',
                               help='analyze content of files')
    github_parser.add_argument('--wide-search', action='store_true',
                               help='analyze all tables or sheets in Office files')

    # sensitibot local
    local_parser = subparsers.add_parser(
        'local', formatter_class=custom_formatters.CustomHelpFormatter, help='analyze local repository')
    local_parser.add_argument('path', type=str, metavar='PATH',
                              const='./', nargs='?', help='The path to analyze')
    local_parser.add_argument('--deep-search', action='store_true',
                              help='analyze content of files')
    local_parser.add_argument('--wide-search', action='store_true',
                              help='analyze all tables or sheets in Office files')

    parser._optionals.title = "optional arguments"
    github_parser._optionals.title = "optional arguments"
    local_parser._optionals.title = "optional arguments"

    result = parser.parse_args(args)

    if result.command == None:
        parser.print_help()
        sys.exit(1)  # exit with non-zero exit code

    return result



import shutil
import tempfile

import pandas as pd
import requests
import sqlalchemy as sa

from reader import columns_reader, headers_reader


def read_access_file(file, deep_search=False, wide_search=False):
    """
    Analyzes the access file.

    Args:
        files (str): The file to analyze.
        deep_search (bool): If true, the content of the files will be analyzed.
        wide_search (bool): If true, all the tables or sheets will be analyzed.

    Returns:
        dict: The result of analyzing the access file.
    """
    new_file = file
    if file.startswith("http"):
        new_file = download_file(file)

    try:
        # Establish a connection to the Access database
        connection_string = (
            r"DRIVER={Microsoft Access Driver (*.mdb, *.accdb)};" +
            r"DBQ={};".format(new_file) +
            r"ExtendedAnsiSQL=1;"
        )
        connection_url = sa.engine.URL.create(
            "access+pyodbc", query={"odbc_connect": connection_string})
        engine = sa.create_engine(connection_url)
    except Exception as e:
        error = {"file": file, "error": str(e)}
        return None, error

    try:
        # Query to retrieve table names
        inspector = sa.inspect(engine)
        table_names = inspector.get_table_names()
    except Exception as e:
        error = {"file": file, "error": str(e)}
        return None, error

    tables_to_read = table_names
    if not wide_search and len(table_names) > 1:
        read_all_sheets = ask_read_all_tables(len(table_names))
        if not read_all_sheets:
            tables_to_read = ask_which_tables(table_names)

    result_file = {"name": file}

    result_tables = []
    for table_name in tables_to_read:

        # Query to retrieve all records from the table
        query = f"SELECT * FROM {table_name}"
        table_data = pd.read_sql_query(query, engine)

        result_table = read_table(table_name, table_data, deep_search)

        if result_table != None:
            result_tables.append(result_table)

    if len(result_tables) != 0:
        result_file["positive_tables"] = result_tables

    if len(result_file) == 1:
        return None, None

    engine.dispose()

    return result_file, None


def read_table(table_name, table_data, deep_search=False):
    result_table = {"name": table_name}

    headers = table_data.columns.values
    result_headers = headers_reader.analize_headers(headers)

    # Only show headers that have errors
    if len(result_headers) != 0:
        result_table["positive_headers"] = result_headers

    # Only analyze columns if deep_search is enabled
    if deep_search:
        result_columns = columns_reader.analize_columns(table_data, headers)

        # Only show columns that have errors
        if len(result_columns) != 0:
            result_table["positive_columns"] = result_columns

    if len(result_table) == 1:
        return None

    return result_table


def ask_read_all_tables(number_of_tables):
    ask_read_all_tables = input(
        f"\n\t\tDo you want to read all ({number_of_tables}) tables? (yes/no): ")
    while ask_read_all_tables.lower() not in ("yes", "no"):
        ask_read_all_tables = input(
            "\t\tPlease enter either 'yes' or 'no': ")
    return ask_read_all_tables.lower() == "yes"


def ask_which_tables(table_names):
    tables_to_read = []

    print("\t\tSelect which tables to read:")

    for table_name in table_names:
        ask_table = input(
            f"\t\t\tRead table {table_name}? (yes/no): ")

        while ask_table.lower() not in ("yes", "no"):
            ask_table = input(
                "\t\t\tPlease enter either 'yes' or 'no': ")

        if ask_table.lower() == "yes":
            tables_to_read.append(table_name)

    return tables_to_read


def download_file(file):
    # Create a temporary file to save the downloaded Access file
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file_path = temp_file.name

    # Download the Access file from GitHub
    response = requests.get(file, stream=True)
    with open(temp_file_path, "wb") as f:
        response.raw.decode_content = True
        shutil.copyfileobj(response.raw, f)

    return temp_file_path



import sys

from cleaner import cleaner
from github import github
from local import local
from reader import reader
from renderer import renderer
from sensitibot import parser


def main():

    args = parser.parse_args()

    files = {}
    name = ""

    if args.command == 'github':
        name = args.user
        try:
            files = github.process_github(
                args.user, args.repository, args.branch, args.token)
        except github.GitHubAPIException as exception:
            print(f'{exception}\n')
            sys.exit(1)  # exit with non-zero exit code

    if args.command == 'local':
        name = "local"
        files = local.process_local(args.path)

    if files == None:
        print("")
        sys.exit(1)  # exit with non-zero exit code

    result = reader.process_files(files, args.deep_search, args.wide_search)
    if result == None:
        sys.exit(1)  # exit with non-zero exit code

    try:
        renderer.show_result_as_text(result, name, args.deep_search)
    except OSError as exception:
        print(f'{exception}\n')

    if args.command == 'local':
        cleaner.process_cleaner(result)

    print("")



import os
import shutil

import pandas as pd
import sqlalchemy as sa


def clean_access_file(file, replace_file=False):
    """
    Cleans access files.

    Args:
        files (dict): The access file to clean.
        replace_files (bool): If true, the file will be replaced by the clean one.

    Returns:
        None.
    """
    filename = file["name"]
    positive_tables = file["positive_tables"] if "positive_tables" in file else [
    ]

    base_filename, extension = os.path.splitext(filename)

    if not replace_file:
        base_filename = base_filename + "_clean"

    new_filename = base_filename + extension

    if not replace_file:
        shutil.copyfile(filename, new_filename)

    try:
        connection_string = (
            r"DRIVER={Microsoft Access Driver (*.mdb, *.accdb)};" +
            r"DBQ={};".format(new_filename) +
            r"ExtendedAnsiSQL=1;"
        )
        connection_url = sa.engine.URL.create(
            "access+pyodbc", query={"odbc_connect": connection_string})
        engine = sa.create_engine(connection_url)
    except Exception as e:
        error = {"file": new_filename, "error": str(e)}

    for positive_table in positive_tables:
        positive_name = positive_table["name"] if "name" in positive_table else ""
        positive_headers = positive_table["positive_headers"] if "positive_headers" in positive_table else [
        ]
        positive_columns = positive_table["positive_columns"] if "positive_columns" in positive_table else dict(
        )

        deleted_columns = []

        for positive_header in positive_headers:
            if positive_header not in deleted_columns:
                delete_column(
                    new_filename, positive_table["name"], positive_header, engine)
                deleted_columns.append(positive_header)

        for positive_column in positive_columns.keys():
            if positive_column not in deleted_columns:
                delete_column(
                    new_filename, positive_table["name"], positive_column, engine)
                deleted_columns.append(positive_column)

    engine.dispose()


def delete_column(filename, table_name, column_name, engine):
    # If the column exists, remove it from the table
    with engine.begin() as connection:
        connection.execute(
            sa.text(f'ALTER TABLE "{table_name}" DROP COLUMN "{column_name}"'))



import pandas as pd

from reader import columns_reader, headers_reader


def read_excel_file(file, deep_search=False, wide_search=False):
    """
    Analyzes the excel file.

    Args:
        files (str): The file to analyze.
        deep_search (bool): If true, the content of the files will be analyzed.
        wide_search (bool): If true, all the tables or sheets will be analyzed.

    Returns:
        dict: The result of analyzing the excel file.
    """
    try:
        data = pd.ExcelFile(file, engine='openpyxl')
    except Exception as e:
        error = {"file": file, "error": str(e)}
        return None, error

    sheet_names = data.sheet_names
    sheets_to_read = sheet_names

    if not wide_search and len(sheet_names) > 1:
        read_all_sheets = ask_read_all_sheets(len(sheet_names))
        if not read_all_sheets:
            sheets_to_read = ask_which_sheets(sheet_names)

    result_file = {"name": file}

    result_sheets = []
    for sheet_name in sheets_to_read:
        sheet_data = data.parse(sheet_name, dtype=str)
        result_sheet = read_sheet(sheet_name, sheet_data, deep_search)

        if result_sheet != None:
            result_sheets.append(result_sheet)

    if len(result_sheets) != 0:
        result_file["positive_sheets"] = result_sheets

    if len(result_file) == 1:
        return None, None

    return result_file, None


def read_sheet(sheet_name, sheet_data, deep_search=False):
    result_sheet = {"name": sheet_name}

    headers = sheet_data.columns.values
    result_headers = headers_reader.analize_headers(headers)

    # Only show headers that have errors
    if len(result_headers) != 0:
        result_sheet["positive_headers"] = result_headers

    # Only analyze columns if deep_search is enabled
    if deep_search:
        result_columns = columns_reader.analize_columns(sheet_data, headers)

        # Only show columns that have errors
        if len(result_columns) != 0:
            result_sheet["positive_columns"] = result_columns

    if len(result_sheet) == 1:
        return None

    return result_sheet


def ask_read_all_sheets(number_of_sheets):
    ask_read_all_sheets = input(
        f"\n\t\tDo you want to read all ({number_of_sheets}) sheets? (yes/no): ")
    while ask_read_all_sheets.lower() not in ("yes", "no"):
        ask_read_all_sheets = input(
            "\t\tPlease enter either 'yes' or 'no': ")
    return ask_read_all_sheets.lower() == "yes"


def ask_which_sheets(sheet_names):
    sheets_to_read = []

    print("\t\tSelect which sheets to read:")

    for sheet_name in sheet_names:
        ask_sheet = input(
            f"\t\t\tRead sheet {sheet_name}? (yes/no): ")

        while ask_sheet.lower() not in ("yes", "no"):
            ask_sheet = input(
                "\t\t\tPlease enter either 'yes' or 'no': ")

        if ask_sheet.lower() == "yes":
            sheets_to_read.append(sheet_name)

    return sheets_to_read
